import json

import requests

from ..exceptions import InternalError


class HttpService:
    """Http Call Header"""
    HEADER = {'Accept': 'application/json', 'Content-type': 'application/json'}

    """
        Generic Http request function

        :param method: 'GET', 'POST' or 'PUT'
        :param url: the requested URL
        :param headers: the headers
        data the request boy, if needed

        :raises InternalError: if status code > 400 or JSON parsing error
        :raises Connection: if connection fails

        :returns: the request's response content
    """
    def request(self, method, url, headers, data=None):
        if False:
            print('\n')
            print(method)
            print(headers)
            print(url)
            print(json.dumps(data))
        try:
            r = requests.request(method=method, url=url, headers=headers, data=data)
            if r.status_code >= 400:
                raise InternalError(r.content)
            content = r.json()
            content['status_code'] = r.status_code
        except json.decoder.JSONDecodeError:
            raise InternalError
        except requests.exceptions.ConnectionError:
            raise ConnectionError
        return content

    """GET function, calls with the right arguments request()"""
    def get(self, url):
        return self.request(method='GET', url=url, headers=self.HEADER)

    """POST function, calls with the right arguments request()"""
    def post(self, url, data=None):
        return self.request(method='POST', url=url, headers=self.HEADER, data=data)

    """PUT function, calls with the right arguments request()"""
    def put(self, url, data=None):
        return self.request(method='PUT', url=url, headers=self.HEADER, data=data)


import time

import pytest

from zenaton.abstracts.task import Task
from zenaton.traits.zenatonable import Zenatonable


class Task0(Task, Zenatonable):

    def handle(self):
        print('Task 0 starts')
        time.sleep(3)
        print('Task 0 ends')
        return 0


@pytest.fixture
def task0():
    return Task0()


class Task1(Task, Zenatonable):

    def handle(self):
        print('Task 1 starts')
        time.sleep(5)
        print('Task 1 ends')
        return 1


@pytest.fixture
def task1():
    return Task1()


class Task2(Task, Zenatonable):

    def handle(self):
        print('Task 2 starts')
        time.sleep(7)
        print('Task 2 ends')
        return 2


@pytest.fixture
def task2():
    return Task2()


class Task3(Task, Zenatonable):

    def handle(self):
        print('Task 3 starts')
        time.sleep(9)
        print('Task 3 ends')
        return 3


@pytest.fixture
def task3():
    return Task3()



import json
import datetime


class Properties:
    SPECIAL_CASES = [
        Exception,
        datetime.datetime,
        datetime.date,
        datetime.time,
    ]

    def blank_instance(self, class_):

        if isinstance(class_, type(None)) or class_ == 'NoneType':
            return None

        class Empty:
            pass

        output = Empty()
        output.__class__ = class_
        return output

    def from_(self, object_):
        if not object_:
            return None
        if self.is_special_case(object_):
            return self.from_complex_type(object_)
        else:
            if hasattr(object_, 'buffer'):
                return object_.buffer
            if hasattr(object_, 'args'):
                return object_.args
            return vars(object_)

    def set(self, object_, properties):
        if properties != (None,) and properties is not None:
            for key, value in properties.items():
                setattr(object_, key, value)
        return object_

    def object_from(self, class_, properties, super_class=None):
        object_ = self.blank_instance(class_)
        self.check_class(object_, super_class)
        self.set(object_, properties)
        return object_

    def check_class(self, object_, super_class):
        error_message = 'Error - #{object.class} should be an instance of #{super_class}'
        if not issubclass(type(object_), super_class):
            raise Exception(error_message)

    def valid_object(self, object_, super_class):
        return not super_class or issubclass(object_, super_class)

    def from_complex_type(self, object_):
        if isinstance(object_, datetime.datetime):
            return {'year': object_.year, 'month': object_.month, 'day': object_.day, 'hour': object_.hour,
                    'minute': object_.minute, 'second': object_.second, 'microsecond': object_.microsecond,
                    'tzinfo': object_.tzinfo}
        if isinstance(object_, datetime.date):
            return {'year': object_.year, 'month': object_.month, 'day': object_.day}
        if isinstance(object_, datetime.time):
            return {'hour': object_.hour, 'minute': object_.minute, 'second': object_.second,
                    'microsecond': object_.microsecond, 'tzinfo': object_.tzinfo}
        if issubclass(type(object_), BaseException):
            return {'error_class': object_.__class__.__name__, 'error_args': list(object_.args)}
        return json.dumps(object_)

    def set_complex_type(self, object_):
        return json.loads(object_)

    def is_special_case(self, object_):
        return type(object_) in self.SPECIAL_CASES or isinstance(object_, BaseException)



import pytest
import datetime


@pytest.mark.usefixtures("wait")
def test_get_duration(wait):

    def assert_duration():
        assert wait.get_duration() == duration

    wait.seconds(20)
    duration = 20.0
    assert_duration()
    wait.seconds(20)
    duration += 20
    assert wait.get_duration() == duration
    wait.hours(10)
    duration += (10 * 3600)
    assert wait.get_duration() == duration
    wait.days(1)
    duration += (24 * 3600)
    assert wait.get_duration() == duration
    wait.weeks(1)
    duration += (7 * 24 * 3600)
    assert wait.get_duration() == duration
    wait.months(1)
    duration28 = duration + (24 * 3600 * 28)
    duration31 = duration + (24 * 3600 * 31)
    duration += (24 * 3600 * 30)
    assert duration28 <= wait.get_duration() <= duration31
    wait.years(1)
    duration_31_full_year = duration31 + (24 * 3600 * 365)
    duration_28_leap_year = duration28 - (24 * 3600)
    assert duration_28_leap_year <= wait.get_duration() <= duration_31_full_year


@pytest.mark.usefixtures("wait")
def test_time_functions(wait):
    for name in ['seconds', 'minutes', 'hours', 'days', 'weeks']:
        getattr(wait, name)(1)
        assert wait.buffer[name] == 1
        getattr(wait, name)(1)
        assert wait.buffer[name] == 2
        wait.buffer = {}
    wait.months(1)
    assert 28 <= wait.buffer['days'] <= 31
    wait.months(1)
    assert 59 <= wait.buffer['days'] <= 62
    wait.buffer = {}
    wait.years(1)
    assert wait.buffer['days'] in [364, 365]
    wait.years(1)
    assert wait.buffer['days'] in [729, 730]


@pytest.mark.usefixtures("wait")
def test_add_months(wait):
    days = wait.months_to_days(1)
    assert 28 <= days <= 31
    days = wait.months_to_days(2)
    assert 56 <= days <= 62
    assert wait.months_to_days(12) == 365


@pytest.mark.usefixtures("wait")
def test_add_years(wait):
    days = wait.years_to_days(1)
    assert days in [364, 365]


@pytest.mark.usefixtures("wait")
def test_init_now_then(wait):
    now, now_dup = wait._WithDuration__init_now_then()
    assert type(now) is datetime.datetime
    assert type(now_dup) is datetime.datetime


@pytest.mark.usefixtures("wait")
def test_push(wait):
    wait.buffer = {}
    wait.seconds(10)
    assert wait.buffer['seconds'] == 10
    wait.seconds(10)
    assert wait.buffer['seconds'] == 20
    wait.hours(20)
    assert wait.buffer['hours'] == 20


@pytest.mark.usefixtures("wait")
def test_apply_duration(wait):
    time = datetime.datetime.now()
    assert (time + datetime.timedelta(seconds=10)) == wait._WithDuration__apply_duration('seconds', 10, time)


@pytest.mark.usefixtures("wait")
def test_diff_in_seconds(wait):
    now = datetime.datetime.now()
    now_plus_10 = now + datetime.timedelta(seconds=10)
    assert wait._WithDuration__diff_in_secondes(now, now_plus_10) == 10



import json
import os
import urllib

from .abstracts.workflow import Workflow
from .exceptions import InvalidArgumentError
from .services.http_service import HttpService
from .services.properties import Properties
from .services.serializer import Serializer
from .singleton import Singleton
from .workflows.version import Version


class Client(metaclass=Singleton):
    ZENATON_API_URL = 'https://api.zenaton.com/v1'  # Zenaton api url
    ZENATON_WORKER_URL = 'http://localhost'  # Default worker url
    DEFAULT_WORKER_PORT = 4001  # Default worker port
    WORKER_API_VERSION = 'v_newton'  # Default worker api version

    MAX_ID_SIZE = 256  # Limit on length of custom ids

    APP_ENV = 'app_env'  # Parameter name for the application environment
    APP_ID = 'app_id'  # Parameter name for the application ID
    API_TOKEN = 'api_token'  # Parameter name for the API token

    ATTR_ID = 'custom_id'  # Parameter name for custom ids
    ATTR_NAME = 'name'  # Parameter name for workflow names
    ATTR_CANONICAL = 'canonical_name' # Parameter name for version name
    ATTR_DATA = 'data'  # Parameter name for json payload
    ATTR_PROG = 'programming_language'  # Parameter name for the language
    ATTR_MODE = 'mode'  # Parameter name for the worker update mode
    ATTR_MAX_PROCESSING_TIME = 'max_processing_time' # Pararameter name for the max processing time

    PROG = 'Python'  # The current programming language

    EVENT_INPUT = 'event_input'  # Parameter name for event data
    EVENT_NAME = 'event_name'  # Parameter name for event name

    WORKFLOW_KILL = 'kill'  # Worker update mode to stop a worker
    WORKFLOW_PAUSE = 'pause'  # Worker udpate mode to pause a worker
    WORKFLOW_RUN = 'run'  # Worker update mode to resume a worker

    def __init__(self, app_id, api_token, app_env):
        self.app_id = app_id
        self.api_token = api_token
        self.app_env = app_env
        self.http = HttpService()
        self.serializer = Serializer()
        self.properties = Properties()

    """
        Gets the url for the workers
        :param String resource the endpoint for the worker
        :param String params url encoded parameters to include in request
        :returns String the workers url with parameters
    """
    def worker_url(self, resource='', params=''):
        base_url = os.environ.get('ZENATON_WORKER_URL') or self.ZENATON_WORKER_URL
        port = os.environ.get('ZENATON_WORKER_PORT') or self.DEFAULT_WORKER_PORT
        url = '{}:{}/api/{}/{}?'.format(base_url, port, self.WORKER_API_VERSION, resource)
        return self.add_app_env(url, params)

    """
        Gets the url for zenaton api
        :param String resource the endpoint for the api
        :param String params url encoded parameters to include in request
        :returns String the api url with parameters
    """
    def website_url(self, resource='', params=''):
        api_url = os.environ.get('ZENATON_API_URL') or self.ZENATON_API_URL
        url = '{}/{}?{}={}&'.format(api_url, resource, self.API_TOKEN, self.api_token)
        return self.add_app_env(url, params)

    def send_event_url(self):
        return self.worker_url('events')

    """
        Start the specified workflow
        :params .abstracts.workflow.Workflow flow
    """
    def start_workflow(self, flow):
        return self.http.post(self.instance_worker_url(),
                              data=json.dumps({
                                  self.ATTR_PROG: self.PROG,
                                  self.ATTR_CANONICAL: self.canonical_name(flow),
                                  self.ATTR_NAME: self.class_name(flow),
                                  self.ATTR_DATA: self.serializer.encode(self.properties.from_(flow)),
                                  self.ATTR_ID: self.parse_custom_id_from(flow)
                       }))

    def start_task(self, task):
        return self.http.post(self.worker_url('tasks'),
                              data=json.dumps({
                                  self.ATTR_PROG: self.PROG,
                                  self.ATTR_NAME: self.class_name(task),
                                  self.ATTR_DATA: self.serializer.encode(self.properties.from_(task)),
                                  self.ATTR_MAX_PROCESSING_TIME: task.max_processing_time() if hasattr(task, 'max_processing_time') else None
                       }))

    def update_instance(self, workflow, custom_id, mode):
        params = '{}={}'.format(self.ATTR_ID, custom_id)
        url = self.instance_worker_url(params)
        options = json.dumps({
            self.ATTR_PROG: self.PROG,
            self.ATTR_NAME: workflow.__name__,
            self.ATTR_MODE: mode
        })
        return self.http.put(url, options)

    """
        Sends an event to a workflow
        :param String workflow_name  the class name of the workflow
        :param String custom_id the custom ID of the workflow (if any)
        :param .abstracts.Event event the event to send
        :returns None
    """
    def send_event(self, workflow_name, custom_id, event):
        body = json.dumps({
            self.ATTR_PROG: self.PROG,
            self.ATTR_NAME: workflow_name,
            self.ATTR_ID: custom_id,
            self.EVENT_NAME: type(event).__name__,
            self.EVENT_INPUT: self.serializer.encode(self.properties.from_(event)),
        })
        return self.http.post(self.send_event_url(), body)

    """
        Finds a workflow
        :param String workflow_name the class name of the workflow
        :param String custom_id the custom ID of the workflow (if any)
        :return .abstracts.workflow.Workflow
    """

    def find_workflow(self, workflow, custom_id):

        params = '{}={}&{}={}&{}={}'.format(
            self.ATTR_ID,
            custom_id,
            self.ATTR_NAME,
            workflow.__name__,
            self.ATTR_PROG,
            self.PROG)
        response = self.http.get(self.instance_website_url(params))

        if response.get('data', None) is not None:
            data = response['data']
            return self.properties.object_from(
                workflow,
                self.serializer.decode(data['properties']),
                Workflow
            )
        else:
            return None

    """
        Stops a workflow
        :param String workflow_name the class name of the workflow
        :param String custom_id the custom ID of the workflow, if any
        :returns None
    """
    def kill_workflow(self, workflow_name, custom_id):
        return self.update_instance(workflow_name, custom_id, self.WORKFLOW_KILL)

    """
        Pauses a workflow
        :param String workflow_name the class name of the workflow
        :param String custom_id the custom ID of the workflow, if any
        :returns None
    """
    def pause_workflow(self, workflow_name, custom_id):
        return self.update_instance(workflow_name, custom_id, self.WORKFLOW_PAUSE)

    """
        Resumes a workflow
        :param String workflow_name the class name of the workflow
        :param String custom_id the custom ID of the workflow, if any
        :returns None
    """
    def resume_workflow(self, workflow_name, custom_id):
        return self.update_instance(workflow_name, custom_id, self.WORKFLOW_RUN)

    def instance_website_url(self, params=''):
        return self.website_url('instances', params)

    def instance_worker_url(self, params=''):
        return self.worker_url('instances', params)

    def add_app_env(self, url, params):
        app_env = '{}={}&'.format(self.APP_ENV, self.app_env) if self.app_env else ''
        app_id = '{}={}&'.format(self.APP_ID, self.app_id) if self.app_id else ''
        return '{}{}{}{}'.format(url, app_env, app_id, urllib.parse.quote_plus(params, safe='=&'))

    def parse_custom_id_from(self, flow):
        custom_id = flow.id()
        if custom_id is not None:
            if not isinstance(custom_id, str) and not isinstance(custom_id, int):
                raise InvalidArgumentError('Provided ID must be a string or an integer')
            custom_id = str(custom_id)
            if len(custom_id) > self.MAX_ID_SIZE:
                raise InvalidArgumentError('Provided Id must not exceed {} bytes'.format(self.MAX_ID_SIZE))
        return custom_id

    def canonical_name(self, flow):
        return type(flow).__name__ if issubclass(type(flow), Version) else None

    def class_name(self, flow):
        if issubclass(type(flow), Version):
            return type(flow.current_implementation()).__name__
        return type(flow).__name__



from .fixtures.fixtures import dummy_object
from .fixtures.fixture_client import client
from .fixtures.fixture_task import task0, task1, task2, task3
from .fixtures.fixture_workflow import sequential_workflow, version_workflow
from .fixtures.fixture_event import my_event
from .fixtures.fixture_engine import engine
from .fixtures.fixture_wait import wait, wait_event
from .fixtures.fixture_serializer import serializer
from .fixtures.fixture_properties import properties



import copy
import datetime
import calendar

import pytz


class WithDuration:

    def get_duration(self):
        if self.buffer is None:
            return
        now, now_dup = self.__init_now_then()
        for (time_unit, time_value) in self.buffer.items():
            now_dup = self.__apply_duration(time_unit, time_value, now_dup)
        return self.__diff_in_secondes(now, now_dup)

    def seconds(self, value):
        self.__push('seconds', value)
        return self

    def minutes(self, value):
        self.__push('minutes', value)
        return self

    def hours(self, value):
        self.__push('hours', value)
        return self

    def days(self, value):
        self.__push('days', value)
        return self

    def weeks(self, value):
        self.__push('weeks', value)
        return self

    # Inspired by https://stackoverflow.com/a/50301887
    def months_to_days(self, months):

        now = datetime.datetime.today().date()
        months_count = now.month + months

        year = now.year + int(months_count / 13)

        month = (months_count % 12)
        if month == 0:
            month = 12

        day = now.day
        last_day_of_month = calendar.monthrange(year, month)[1]
        if day > last_day_of_month:
            day = last_day_of_month

        new_date = datetime.date(year, month, day)

        return (new_date - now).days

    def months(self, value):
        days = self.months_to_days(value)
        self.__push('days', days)
        return self

    def years_to_days(self, years):
        return self.months_to_days(years * 12)

    def years(self, value):
        days = self.years_to_days(value)
        self.__push('days', days)
        return self

    def __init_now_then(self):
        tz = self.timezone if self.timezone else 'UTC'
        now = datetime.datetime.now(pytz.timezone(tz))
        return now, copy.deepcopy(now)

    def __push(self, key, value=1):
        if not hasattr(self, 'buffer'):
            self.buffer = {}
        if not self.buffer.get(key):
            self.buffer[key] = value
        else:
            self.buffer[key] += value

    def __apply_duration(self, time_unit, time_value, time):
        args = {str(time_unit): int(time_value)}
        return time + datetime.timedelta(**args)

    def __diff_in_secondes(self, before, after):
        return (after - before) / datetime.timedelta(seconds=1)



import pytest

from zenaton.tasks.wait import Wait
from .fixture_event import MyEvent


@pytest.fixture
def wait():
    wait = Wait()
    wait.set_timezone('Europe/Paris')
    return wait


@pytest.fixture
def wait_event():
    return Wait(MyEvent)



import datetime
import itertools

import pytz

from .with_duration import WithDuration
from ..exceptions import ExternalError, InternalError


class WithTimestamp(WithDuration):
    WEEKDAYS = ('monday', 'tuesday', 'wednesday', 'thursday', 'friday', 'saturday', 'sunday')
    MODE_AT = 'AT'  # When specifying a time
    MODE_WEEK_DAY = 'WEEK_DAY'  # When specifying a day of the week
    MODE_MONTH_DAY = 'MONTH_DAY'  # When specifying a day of the month
    MODE_TIMESTAMP = 'TIMESTAMP'  # When specifying a unix timestamp

    timezone = 'UTC'

    HAS_DAY = False

    def get_timetamp_or_duration(self):
        if getattr(self, 'buffer', None) is None:
            return [None, None]
        now, now_dup = self._WithDuration__init_now_then()
        self.mode = None
        for (time_unit, time_value) in sorted(self.buffer.items(), key=lambda kv: kv[0]):
            now_dup = self.__apply(time_unit, time_value, now, now_dup)
        if self.mode is None:
            return [None, self._WithDuration__diff_in_secondes(now, now_dup)]
        if self.mode == self.MODE_TIMESTAMP:
            return [int(now_dup), None]
        return [int(now_dup.timestamp()), None]

    def timestamp(self, value):
        self._WithDuration__push('timestamp', value)
        return self

    def at(self, value):
        self._WithDuration__push('at', value)
        return self

    def day_of_month(self, value):
        self._WithDuration__push('day_of_month', value)
        return self

    def monday(self, value=1):
        self.HAS_DAY = True
        self._WithDuration__push('monday', value)
        return self

    def tuesday(self, value=1):
        self.HAS_DAY = True
        self._WithDuration__push('tuesday', value)
        return self

    def wednesday(self, value=1):
        self.HAS_DAY = True
        self._WithDuration__push('wednesday', value)
        return self

    def thursday(self, value=1):
        self.HAS_DAY = True
        self._WithDuration__push('thursday', value)
        return self

    def friday(self, value=1):
        self.HAS_DAY = True
        self._WithDuration__push('friday', value)
        return self

    def saturday(self, value=1):
        self.HAS_DAY = True
        self._WithDuration__push('saturday', value)
        return self

    def sunday(self, value=1):
        self.HAS_DAY = True
        self._WithDuration__push('sunday', value)
        return self

    def __apply(self, method, value, now, now_dup):
        if method in self.WEEKDAYS:
            return self.__weekday(value, method, now_dup)
        elif method == 'timestamp':
            return self.__timestamp(value)
        elif method == 'at':
            return self.__at(value, now, now_dup)
        elif method == 'day_of_month':
            return self.__day_of_month(value, now, now_dup)
        else:
            return self._WithDuration__apply_duration(method, value, now)

    def __weekday(self, value, day, now_dup):
        self.__set_mode(self.MODE_WEEK_DAY)
        for _ in itertools.repeat(None, value):
            now_dup = self.__next_weekday(now_dup, day)
        return now_dup

    # https://stackoverflow.com/questions/6558535/
    def __next_weekday(self, d, weekday):
        weekday = self.WEEKDAYS.index(weekday)
        days_ahead = weekday - d.weekday()
        if days_ahead <= 0:  # Target day already happened this week
            days_ahead += 7
        return d + datetime.timedelta(days_ahead)

    def __timestamp(self, timestamp):
        self.__set_mode(self.MODE_TIMESTAMP)
        return timestamp

    def __at(self, time, now, now_dup):
        self.__set_mode(self.MODE_AT)
        hour, minute, second = time.split(':')
        hour, minute, second = int(hour), int(minute), int(second)
        now_dup = now_dup.replace(hour=hour, minute=minute, second=second)
        if now > now_dup:
            now_dup += self.__delay()
        elif self.HAS_DAY:
            now_dup -= datetime.timedelta(weeks=1)
        return now_dup

    def __delay(self):
        if self.mode == self.MODE_AT:
            return datetime.timedelta(days=1)
        elif self.mode == self.MODE_WEEK_DAY:
            return datetime.timedelta(weeks=1)
        elif self.mode == self.MODE_MONTH_DAY:
            days = self.months_to_days(1)
            return datetime.timedelta(days=days)
        else:
            raise InternalError('Unknown mode: {}'.format(self.mode))

    def __day_of_month(self, day, now, now_dup):
        self.__set_mode(self.MODE_MONTH_DAY)
        now_dup = now_dup.replace(day=day)
        if now >= now_dup:
            if now_dup.month > 11:
                now_dup = now_dup.replace(month=1, year=now_dup.year + 1)
            else:
                now_dup = now_dup.replace(month=now_dup.month + 1)
        return now_dup

    def __set_mode(self, mode):
        if not hasattr(self, 'mode') or self.mode is None:
            self.mode = mode
            return
        if mode == self.mode or self.__is_timestamp_mode_set(mode):
            raise ExternalError('Incompatible definition in Wait methods')
        else:
            self.mode = self.MODE_AT

    def __is_timestamp_mode_set(self, mode):
        return (self.mode is not None and self.MODE_TIMESTAMP == mode) or self.mode == self.MODE_TIMESTAMP

    def set_timezone(self, timezone):
        if not self.__is_valid_timezone(timezone):
            raise ExternalError('Unknown timezone')
        self.__class__.timezone = timezone
        return self

    def __is_valid_timezone(self, timezone):
        return timezone in pytz.all_timezones



import pytest
import json
import datetime


@pytest.mark.usefixtures("serializer")
def test_str(serializer):
    str_ = 'foo bar'
    assert '{"d": "foo bar", "s": []}' == serializer.encode(str_)
    assert serializer.decode('{"d": "foo bar", "s": []}') == str_

    empty_str_ = ''
    assert '{"d": "", "s": []}' == serializer.encode(empty_str_)
    assert serializer.decode('{"d": "", "s": []}') == empty_str_


@pytest.mark.usefixtures("serializer")
def test_int(serializer):
    int_ = 10
    assert '{"d": 10, "s": []}' == serializer.encode(int_)
    assert serializer.decode('{"d": 10, "s": []}') == int_

    neg_int_ = -10
    assert '{"d": -10, "s": []}' == serializer.encode(neg_int_)
    assert serializer.decode('{"d": -10, "s": []}') == neg_int_

    zero_int_ = 0
    assert '{"d": 0, "s": []}' == serializer.encode(zero_int_)
    assert serializer.decode('{"d": 0, "s": []}') == zero_int_


@pytest.mark.usefixtures("serializer")
def test_float(serializer):
    float_ = 10.0
    assert '{"d": 10.0, "s": []}' == serializer.encode(float_)
    assert serializer.decode('{"d": 10.0, "s": []}') == float_

    neg_float_ = -10.0
    assert '{"d": -10.0, "s": []}' == serializer.encode(neg_float_)
    assert serializer.decode('{"d": -10.0, "s": []}') == neg_float_

    zero_float_ = 0.0
    assert '{"d": 0.0, "s": []}' == serializer.encode(zero_float_)
    assert serializer.decode('{"d": 0.0, "s": []}') == zero_float_


@pytest.mark.usefixtures("serializer")
def test_bool(serializer):
    true = True
    assert '{"d": true, "s": []}' == serializer.encode(true)
    assert serializer.decode('{"d": true, "s": []}') == true

    false = False
    assert '{"d": false, "s": []}' == serializer.encode(false)
    assert serializer.decode('{"d": false, "s": []}') == false


@pytest.mark.usefixtures("serializer")
def test_none(serializer):
    none = None
    assert '{"d": null, "s": []}' == serializer.encode(none)
    assert serializer.decode('{"d": null, "s": []}') == none


datetime_ = datetime.datetime(2018, 10, 16, 16, 53, 29, 164069)


@pytest.mark.usefixtures("serializer")
def test_datetime(serializer):
    assert '{"o": "@zenaton#0",' \
           ' "s": [{"n": "datetime",' \
           ' "p": {"day": 16, "hour": 16, "microsecond": 164069, "minute": 53, "month": 10, "second": 29, "tzinfo": null, "year": 2018}}]}' == serializer.encode(
        datetime_)


@pytest.mark.usefixtures("serializer")
def test_date(serializer):
    date_ = datetime_.date()
    assert '{"o": "@zenaton#0", "s": [{"n": "date", "p": {"day": 16, "month": 10, "year": 2018}}]}' == serializer.encode(
        date_)


@pytest.mark.usefixtures("serializer")
def test_time(serializer):
    time_ = datetime_.time()
    assert '{"o": "@zenaton#0",' \
           ' "s": [{"n": "time", "p": {"hour": 16, "microsecond": 164069, "minute": 53, "second": 29, "tzinfo": null}}]}' == serializer.encode(
        time_)


@pytest.mark.usefixtures("serializer")
def test_dict(serializer):
    dict_ = {'foo': 'bar'}
    assert '{"o": "@zenaton#0", "s": [{"a": {"foo": "bar"}}]}' == serializer.encode(dict_)
    assert serializer.decode('{"o": "@zenaton#0", "s": [{"a": {"foo": "bar"}}]}') == dict_

    empty_dict_ = dict()
    assert '{"o": "@zenaton#0", "s": [{"a": {}}]}' == serializer.encode(empty_dict_)
    assert serializer.decode('{"o": "@zenaton#0", "s": [{"a": {}}]}') == empty_dict_


@pytest.mark.usefixtures("serializer")
def test_list(serializer):
    list_ = ['foo', 'bar', 1, 10]
    assert '{"o": "@zenaton#0", "s": [{"a": ["foo", "bar", 1, 10]}]}' == serializer.encode(list_)
    assert serializer.decode('{"o": "@zenaton#0", "s": [{"a": ["foo", "bar", 1, 10]}]}') == list_

    empty_list_ = []
    assert '{"o": "@zenaton#0", "s": [{"a": []}]}' == serializer.encode(empty_list_)
    assert serializer.decode('{"o": "@zenaton#0", "s": [{"a": []}]}') == empty_list_


dict0 = {}
dict1 = {'dict0': dict0}
dict0['dict1'] = dict1


@pytest.mark.usefixtures("serializer")
def test_circular_dict(serializer):
    assert '{"o": "@zenaton#0",' \
           ' "s": [{"a": {"dict1": "@zenaton#1"}}, {"a": {"dict0": "@zenaton#0"}}]}' == serializer.encode(dict0)
    decoded = serializer.decode('{"o": "@zenaton#0",'
                                ' "s": [{"a": {"dict1": "@zenaton#1"}}, {"a": {"dict0": "@zenaton#0"}}]}')
    assert id(decoded) == id(decoded['dict1']['dict0'])
    assert id(decoded['dict1']) == id(decoded['dict1']['dict0']['dict1'])

    assert '{"o": "@zenaton#0",' \
           ' "s": [{"a": {"dict0": "@zenaton#1"}}, {"a": {"dict1": "@zenaton#0"}}]}' == serializer.encode(dict1)
    decoded = serializer.decode('{"o": "@zenaton#0",'
                                ' "s": [{"a": {"dict0": "@zenaton#1"}}, {"a": {"dict1": "@zenaton#0"}}]}')
    assert id(decoded) == id(decoded['dict0']['dict1'])
    assert id(decoded['dict0']) == id(decoded['dict0']['dict1']['dict0'])


list0 = ['foo']
list1 = ['bar']
list0.append(list1)
list1.append(list0)


@pytest.mark.usefixtures("serializer")
def test_circular_list(serializer):
    assert '{"o": "@zenaton#0",' \
           ' "s": [{"a": ["foo", "@zenaton#1"]}, {"a": ["bar", "@zenaton#0"]}]}' == serializer.encode(list0)
    decoded = serializer.decode('{"o": "@zenaton#0",'
                                ' "s": [{"a": ["foo", "@zenaton#1"]}, {"a": ["bar", "@zenaton#0"]}]}')
    assert id(decoded) == id(decoded[1][1])
    assert id(decoded[1]) == id(decoded[1][1][1])

    assert '{"o": "@zenaton#0",' \
           ' "s": [{"a": ["bar", "@zenaton#1"]}, {"a": ["foo", "@zenaton#0"]}]}' == serializer.encode(list1)
    decoded = serializer.decode('{"o": "@zenaton#0",'
                                ' "s": [{"a": ["bar", "@zenaton#1"]}, {"a": ["foo", "@zenaton#0"]}]}')
    assert id(decoded) == id(decoded[1][1])
    assert id(decoded[1]) == id(decoded[1][1][1])


class Container:
    pass


container = Container()
container.dict0 = dict0
container.dict1 = dict1


@pytest.mark.usefixtures("serializer")
def test_container(serializer):
    assert ('{"o": "@zenaton#0",' \
            ' "s": [{"n": "Container",' \
            ' "p": {"dict0": "@zenaton#2", "dict1": "@zenaton#1"}},' \
           ' {"a": {"dict0": "@zenaton#2"}},' \
            ' {"a": {"dict1": "@zenaton#1"}}]}' == serializer.encode(container) or
            '{"o": "@zenaton#0",' \
            ' "s": [{"n": "Container",' \
            ' "p": {"dict0": "@zenaton#1", "dict1": "@zenaton#2"}},' \
            ' {"a": {"dict1": "@zenaton#2"}},' \
            ' {"a": {"dict0": "@zenaton#1"}}]}' == serializer.encode(container))


container0 = Container()
container1 = Container()
container0.teammate = container1
container1.teammate = container0


@pytest.mark.usefixtures("serializer")
def test_circular_container(serializer):
    assert '{"o": "@zenaton#0",' \
           ' "s": [{"n": "Container",' \
           ' "p": {"teammate": "@zenaton#1"}},' \
           ' {"n": "Container",' \
           ' "p": {"teammate": "@zenaton#0"}}]}' == serializer.encode(container0)



import copy
import json
import inspect
import importlib.util

from zenaton.services.properties import Properties
from zenaton.exceptions import InvalidArgumentError


class Serializer:
    # this string prefixes ids that are used to identify objects
    ID_PREFIX = '@zenaton#'

    KEY_OBJECT = 'o'  # JSON key for objects
    KEY_OBJECT_NAME = 'n'  # JSON key for class name
    KEY_OBJECT_PROPERTIES = 'p'  # JSON key for object ivars
    KEY_ARRAY = 'a'  # JSON key for array and hashes
    KEY_DATA = 'd'  # JSON key for json compatibles types
    KEY_STORE = 's'  # JSON key for deserialized complex object

    def __init__(self, boot=None, name=None):
        self.properties = Properties()
        self.encoded = []
        self.decoded = []
        self.boot = boot
        self.name = name

    def encode(self, data):
        self.encoded = []
        self.decoded = []
        value = dict()
        if self.__is_basic_type(data):
            value[self.KEY_DATA] = data
        else:
            value[self.KEY_OBJECT] = self.__encode_to_store(data)
        value[self.KEY_STORE] = self.encoded
        return json.dumps(value, sort_keys=True)

    def decode(self, json_string):
        parsed_json = json.loads(json_string)
        encoded_json = copy.deepcopy(parsed_json)
        self.encoded = encoded_json[self.KEY_STORE]
        del encoded_json[self.KEY_STORE]
        self.decoded = []
        if self.KEY_DATA in parsed_json:
            return parsed_json[self.KEY_DATA]
        if self.KEY_ARRAY in parsed_json:
            return self.__decode_enumerable(parsed_json[self.KEY_ARRAY])
        if self.KEY_OBJECT in parsed_json:
            id_ = int(parsed_json[self.KEY_OBJECT][len(self.ID_PREFIX):])
            return self.__decode_from_store(id_, self.encoded[id_])

    def __encode_value(self, value):
        if self.__is_basic_type(value):
            return value
        else:
            return self.__encode_to_store(value)

    def __encode_to_store(self, object_):

        for index, element in enumerate(self.decoded):
            if id(element) == id(object_):
                return self.__store_id(index)

        id_ = len(self.decoded)
        self.insert_at_index(self.decoded, id_, object_)
        self.insert_at_index(self.encoded, id_, self.__encode_object_by_type(object_))
        return self.__store_id(id_)

    def insert_at_index(self, list_, index, value):
        try:
            list_[index] = value
        except IndexError:
            for i in range(0, index - len(list_) + 1):
                list_.append(None)
            list_[index] = value

    def append_at_index(self, list_, index, value):
        if len(list_[index]) >= len(value):
            return
        try:
            list_[index].extend(value)
        except (TypeError, IndexError, AttributeError):
            for i in range(0, index - len(list_) + 1):
                list_.append(None)
            list_[index] = value

    def update_at_index(self, list_, index, value):
        if len(list_[index]) >= len(value):
            return
        try:
            list_[index].update(value)
        except (TypeError, IndexError, AttributeError):
            for i in range(0, index - len(list_) + 1):
                list_.append(None)
            list_[index] = value

    def __encode_object_by_type(self, object_):
        if isinstance(object_, list):
            return self.__encode_list(object_)
        if isinstance(object_, dict):
            return self.__encode_dict(object_)
        return self.__encode_object(object_)

    def __encode_object(self, object_):
        return {
            self.KEY_OBJECT_NAME: type(object_).__name__,
            self.KEY_OBJECT_PROPERTIES: self.__encode_legacy_dict(self.properties.from_(object_))
        }

    def __encode_list(self, list_):
        return {
            self.KEY_ARRAY: [self.__encode_value(element) for element in list_]
        }

    def __encode_dict(self, dict_):
        return {
            self.KEY_ARRAY: {key: self.__encode_value(value) for key, value in dict_.items()}
        }

    def __encode_legacy_dict(self, dict_):
        return {key: self.__encode_value(value) for key, value in dict_.items()}

    def __decode_element(self, value):
        if self.__is_store_id(value):
            id_ = int(value[len(self.ID_PREFIX):])
            encoded = self.encoded[id_]
            return self.__decode_from_store(id_, encoded)
        elif isinstance(value, list):
            return self.__decode_legacy_list(value)
        elif isinstance(value, dict):
            return self.__decode_legacy_dict(value)
        else:
            return value

    def __decode_enumerable(self, enumerable):
        if isinstance(enumerable, list):
            return self.__decode_legacy_list(enumerable)
        if isinstance(enumerable, dict):
            return self.__decode_legacy_dict(enumerable)
        raise InvalidArgumentError('Unknown type')

    def __decode_legacy_list(self, list_):
        return [self.__decode_element(element) for element in list_]

    def __decode_legacy_dict(self, dict_):
        ret = {key: self.__decode_element(value) for key, value in dict_.items()}
        return ret

    def __decode_list(self, id_, list_):
        self.insert_at_index(self.decoded, id_, [])
        decoded_list = [self.__decode_element(element) for element in list_]
        self.append_at_index(self.decoded, id_, decoded_list)
        return self.decoded[id_]

    def __decode_dict(self, id_, dict_):
        self.insert_at_index(self.decoded, id_, dict())
        decoded_dict = {key: self.__decode_element(value) for key, value in dict_.items()}
        self.update_at_index(self.decoded, id_, decoded_dict)
        return self.decoded[id_]

    def __decode_from_store(self, id_, encoded):
        if len(self.decoded) >= id_ + 1 and self.decoded[id_] is not None:
            decoded = self.decoded[id_]
            return decoded
        else:
            encoded_value = encoded.get(self.KEY_ARRAY, None)
            if isinstance(encoded_value, list):
                return self.__decode_list(id_, encoded_value)
            if isinstance(encoded_value, dict):
                return self.__decode_dict(id_, encoded_value)
            return self.__decoded_object(id_, encoded)

    def __decoded_object(self, id_, encoded_object):
        if len(self.decoded) >= id_ + 1 and self.decoded[id_] is not None:
            return self.decoded[id_]
        try:
            object_class = self.import_class(self.name, encoded_object[self.KEY_OBJECT_NAME])
            object_ = self.properties.blank_instance(object_class)
            self.insert_at_index(self.decoded, id_, object_)
            properties = self.__decode_legacy_dict(encoded_object.get(self.KEY_OBJECT_PROPERTIES, None))
            self.properties.set(object_, properties)
            return object_
        except TypeError:
            properties = self.__decode_legacy_dict(encoded_object.get(self.KEY_OBJECT_PROPERTIES, None))
            object_ = object_class(**properties)
            self.insert_at_index(self.decoded, id_, object_)
            return object_
        except KeyError:
            return None

    def __is_store_id(self, string_):
        try:
            suffix = int(string_[len(self.ID_PREFIX):])
        except (TypeError, ValueError):
            return False
        return isinstance(string_, str) and \
               string_.startswith(self.ID_PREFIX) and \
               suffix <= len(self.encoded)

    def __store_id(self, id):
        return '{}{}'.format(self.ID_PREFIX, id)

    def __is_basic_type(self, data):
        return (isinstance(data, str) or
                isinstance(data, int) or
                isinstance(data, float) or
                isinstance(data, bool) or data is None)

    def import_class(self, workflow_name, class_name):
        spec = importlib.util.spec_from_file_location('boot', self.boot)
        boot = importlib.util.module_from_spec(spec)
        spec.loader.exec_module(boot)
        workflow_class = getattr(boot, workflow_name)
        workflow_path = inspect.getfile(workflow_class)
        workflow_spec = importlib.util.spec_from_file_location('workflow', workflow_path)
        workflow_module = importlib.util.module_from_spec(workflow_spec)
        workflow_spec.loader.exec_module(workflow_module)
        return getattr(workflow_module, class_name)



import pytest
import datetime
import copy

from zenaton.tasks.wait import Wait
from zenaton.exceptions import ExternalError

from zenaton.traits.with_timestamp import WithTimestamp


@pytest.mark.usefixtures("wait")
def test_get_timetamp_or_duration_next_weekday(wait):
    wait2 = copy.deepcopy(wait)
    now = datetime.datetime.now()
    now_day = now.weekday() + 1
    weekday = WithTimestamp.WEEKDAYS[now.weekday()]
    getattr(wait2, weekday)()
    assert wait2.get_timetamp_or_duration()[0] - int(now.timestamp()) == (7 * 24 * 60 * 60)
    for i in range(1, 6):
        new_wait = copy.deepcopy(wait)
        getattr(new_wait, new_wait.WEEKDAYS[now_day])(i)
        assert new_wait.get_timetamp_or_duration()[0] - int(now.timestamp()) == (24 * 60 * 60) * (1 + (7 * (i - 1)))


@pytest.mark.usefixtures("wait")
def test_get_timetamp_or_duration_weekday(wait):
    wait2 = copy.deepcopy(wait)
    now = datetime.datetime.now()
    weekday = WithTimestamp.WEEKDAYS[now.weekday()]
    getattr(wait, weekday)().at("{}:{}:{}".format(now.hour + 1, now.minute, now.second))
    assert wait.get_timetamp_or_duration()[0] - int(now.timestamp()) == (60 * 60)
    getattr(wait2, weekday)().at("{}:{}:{}".format(now.hour - 1, now.minute, now.second))
    assert wait2.get_timetamp_or_duration()[0] - int(now.timestamp()) == (7 * 24 * 60 * 60) - (60 * 60)


@pytest.mark.usefixtures("wait")
def test_get_timetamp_or_duration_day_of_month(wait):
    wait2 = copy.deepcopy(wait)
    wait3 = copy.deepcopy(wait)
    now = datetime.datetime.now()
    wait.day_of_month(now.day + 1)
    assert wait.get_timetamp_or_duration()[0] - int(now.timestamp()) == 24 * 60 * 60
    wait2.day_of_month(now.day - 1)
    wait_duration = wait2.get_timetamp_or_duration()[0] - int(now.timestamp())
    assert 27 * 24 * 60 * 60 <= wait_duration <= 30 * 24 * 60 * 60
    wait3.day_of_month(now.day)
    wait_duration = wait3.get_timetamp_or_duration()[0] - int(now.timestamp())
    assert 28 * 24 * 60 * 60 <= wait_duration <= 31 * 24 * 60 * 60


@pytest.mark.usefixtures("wait")
def test_get_timetamp_or_duration_day_of_month_at(wait):
    wait2 = copy.deepcopy(wait)
    now = datetime.datetime.now()
    hour = now.hour
    wait.day_of_month(now.day).at("{}:{}:{}".format(hour + 1, now.minute, now.second))
    assert wait.get_timetamp_or_duration()[0] - int(now.timestamp()) == 60 * 60
    wait2.day_of_month(now.day).at("{}:{}:{}".format(hour - 1, now.minute, now.second))
    assert 28 * 24 * 60 * 60 <= wait2.get_timetamp_or_duration()[0] - int(now.timestamp()) <= 31 * 24 * 60 * 60



@pytest.mark.usefixtures("wait")
def test_get_timetamp_or_duration_at(wait):
    now = datetime.datetime.now()
    wait.set_timezone('Europe/Paris')
    wait.at('{}:{}:{}'.format(now.hour + 1, now.minute, now.second))
    assert wait.get_timetamp_or_duration()[0] - int(now.timestamp()) == 60 * 60


@pytest.mark.usefixtures("wait")
def test_get_timetamp_or_duration_timestamp(wait):
    now_timestamp = int(datetime.datetime.now().timestamp())
    wait.timestamp(now_timestamp + 10.0)
    assert wait.get_timetamp_or_duration()[0] - now_timestamp == 10


@pytest.mark.usefixtures("wait")
def test_get_timetamp_push(wait):
    for name in ['timestamp', 'at', 'day_of_month', 'monday', 'tuesday', 'wednesday', 'thursday', 'friday', 'saturday',
                 'sunday']:
        getattr(wait, name)(1)
        assert wait.buffer.get(name, None)


@pytest.mark.usefixtures("wait")
def test_set_mode(wait):
    wait._WithTimestamp__set_mode(wait.MODE_WEEK_DAY)
    assert wait.mode == wait.MODE_WEEK_DAY
    with pytest.raises(ExternalError):
        wait._WithTimestamp__set_mode(wait.MODE_WEEK_DAY)
    wait._WithTimestamp__set_mode(wait.MODE_MONTH_DAY)
    assert wait.mode == wait.MODE_AT


@pytest.mark.usefixtures("wait")
def test_is_timestamp_mode_set(wait):
    wait.mode = wait.MODE_WEEK_DAY
    assert not wait._WithTimestamp__is_timestamp_mode_set(wait.MODE_WEEK_DAY)
    assert wait._WithTimestamp__is_timestamp_mode_set(wait.MODE_TIMESTAMP)
    assert not wait._WithTimestamp__is_timestamp_mode_set(wait.MODE_WEEK_DAY)


@pytest.mark.usefixtures("wait")
def test_set_timezone(wait):
    wait.set_timezone('Africa/Libreville')
    assert wait.timezone == 'Africa/Libreville'
    new_wait = Wait()
    assert new_wait.timezone == 'Africa/Libreville'
    with pytest.raises(ExternalError):
        wait.set_timezone('Africa/Leconi')


@pytest.mark.usefixtures("wait")
def test_valid_timezone(wait):
    assert wait._WithTimestamp__is_valid_timezone('Africa/Libreville')
    assert not wait._WithTimestamp__is_valid_timezone('Africa/Mvengué')



import pytest

from zenaton.engine import Engine
from zenaton.client import Client


import os

import pytest
from dotenv import load_dotenv

from zenaton.client import Client

# LOADING CONFIG FROM .env file
load_dotenv()
app_id = os.getenv('ZENATON_APP_ID')
api_token = os.getenv('ZENATON_API_TOKEN')
app_env = os.getenv('ZENATON_APP_ENV')

if not app_id:
    raise Exception('Please include your ZENATON_APP_ID in the .env file')

if not api_token:
    raise Exception('Please include your ZENATON_API_TOKEN in the .env file')

if not app_env:
    raise Exception('Please include your ZENATON_APP_ENV in the .env file')

@pytest.fixture
def engine():
    Client(app_id, api_token, app_env)
    return Engine()



import pytest

from zenaton.exceptions import InvalidArgumentError
from zenaton.client import Client

from .utils import validate_url


@pytest.mark.usefixtures("client")
def test_url_functions(client):
    assert validate_url(client.worker_url())
    assert validate_url(client.send_event_url())
    assert validate_url(client.instance_website_url())
    assert validate_url(client.instance_worker_url())


@pytest.mark.usefixtures("client", "sequential_workflow")
def test_class_name(client, sequential_workflow):
    assert type(client.class_name(sequential_workflow)) == str


@pytest.mark.usefixtures("client", "sequential_workflow", "version_workflow")
def test_canonical_name(client, sequential_workflow, version_workflow):
    assert client.canonical_name(sequential_workflow) is None
    assert type(client.canonical_name(version_workflow)) == str


@pytest.mark.usefixtures("client", "sequential_workflow")
def test_parse_custom_id_from(client, sequential_workflow):
    assert type(client.parse_custom_id_from(sequential_workflow)) == str
    sequential_workflow.set_id(0)
    assert type(client.parse_custom_id_from(sequential_workflow)) == str
    sequential_workflow.set_id({})
    with pytest.raises(InvalidArgumentError):
        client.parse_custom_id_from(sequential_workflow)
    with pytest.raises(InvalidArgumentError):
        sequential_workflow.set_id('A' * (Client.MAX_ID_SIZE + 1))
        client.parse_custom_id_from(sequential_workflow)


@pytest.mark.usefixtures("client", "sequential_workflow")
def test_workflow_lifecycle(client, sequential_workflow):
    response = client.start_workflow(sequential_workflow)
    assert response['status_code'] == 201
    response = client.pause_workflow(type(sequential_workflow), sequential_workflow.id())
    assert response['status_code'] == 200
    response = client.resume_workflow(type(sequential_workflow), sequential_workflow.id())
    assert response['status_code'] == 200
    worfklow = client.find_workflow(type(sequential_workflow), sequential_workflow.id())
    assert type(worfklow) == type(sequential_workflow)
    assert worfklow.id() == sequential_workflow.id()
    response = client.kill_workflow(type(sequential_workflow), sequential_workflow.id())
    assert response['status_code'] == 200


@pytest.mark.usefixtures("client", "sequential_workflow", "my_event")
def test_event_workflow_lifecycle(client, sequential_workflow, my_event):
    sequential_workflow.set_id('MyEventWorkflow')
    response = client.start_workflow(sequential_workflow)
    assert response['status_code'] == 201
    response = client.send_event(type(sequential_workflow).__name__, sequential_workflow.id(), my_event)
    assert response['status_code'] == 200
    response = client.kill_workflow(type(sequential_workflow), sequential_workflow.id())
    assert response['status_code'] == 200


def test_url_params_encoding(client):
    assert client.add_app_env('', 'workflow_id').endswith('workflow_id')
    assert client.add_app_env('', 'yann+1@zenaton.com').endswith('&yann%2B1%40zenaton.com')


import pytest

from zenaton.tasks.wait import Wait
from zenaton.exceptions import ExternalError
from ..fixtures.fixture_event import MyEvent


def test_init():
    with pytest.raises(ExternalError):
        Wait(1)


@pytest.mark.usefixtures("wait")
def test_has_handle(wait):
    assert hasattr(wait, 'handle')


@pytest.mark.usefixtures("wait", "wait_event", "my_event")
def test_valid_param(wait, my_event):
    assert not wait.valid_param(1)
    assert wait.valid_param(None)
    assert wait.valid_param("foo")
    assert wait.valid_param(MyEvent)


@pytest.mark.usefixtures("wait", "my_event")
def test_event_class(wait, my_event):
    assert not wait.event_class("foo")
    assert wait.event_class(MyEvent)



from zenaton.singleton import Singleton


class SingleClass(metaclass=Singleton):
    def __init__(self):
        self.foo = 0


def test_unicity():
    SingleClass().foo = 1
    assert SingleClass().foo == 1



import pytest

from .fixture_task import task0, task1, task2, task3
from .fixture_event import MyEvent

from zenaton.abstracts.workflow import Workflow
from zenaton.traits.zenatonable import Zenatonable
from zenaton.workflows.version import Version


class SequentialWorkflow(Workflow, Zenatonable):

    def __init__(self):
        self.id_ = 'MyTestWorkflowId'

    def handle(self):

        a = task0.execute()

        if a > 0:
            task1.execute()
        else:
            task2.execute()

        task3.execute()

    def id(self):
        return self.id_

    def set_id(self, id_):
        self.id_ = id_

    def on_event(self, event):
        if issubclass(MyEvent, type(event)):
            return True
        else:
            return False



@pytest.fixture
def sequential_workflow():
    return SequentialWorkflow()


class VersionWorkflow(Version):

    def versions(self):
        return [SequentialWorkflow, ]


@pytest.fixture
def version_workflow():
    return VersionWorkflow()



from ..abstracts.workflow import Workflow
from ..exceptions import ExternalError


class Builder:
    """Wrapper class around the client to interact with workflows by id"""

    def __init__(self, class_):
        from ..client import Client
        self.check_class(class_)
        self.class_ = class_
        self.client = Client()

    """
        Sets the id of the workflow we want to find
        :param String or None id the id
        :returns Builder the current builder
    """
    def where_id(self, id):
        self.id = id
        return self

    """
        Finds a workflow
        returns Workflow
    """
    def find(self):
        return self.client.find_workflow(self.class_, self.id)

    """
        Sends an event to a workflow
        :param abstracts.event.Event event the event to send
        :returns query.builder the current builder
    """
    def send_event(self, event):
        self.client.send_event(self.class_.__name__, self.id, event)

    """
        Kills a workflow
        :returns query.builder.Builder the current builder
    """
    def kill(self):
        self.client.kill_workflow(self.class_, self.id)
        return self

    """
        Pauses a workflow
        :returns query.builder.Builder the current builder
    """
    def pause(self):
        self.client.pause_workflow(self.class_, self.id)
        return self

    """
        Resumes a workflow
        :returns query.builder.Builder the current builder
    """
    def resume(self):
        self.client.resume_workflow(self.class_, self.id)
        return self

    """
        Checks if class_ is subclass of Workflow
        :param class class_
    """

    def check_class(self, class_):
        msg = '{} should be a subclass of .abstracts.workflow'.format(class_)
        if not issubclass(class_, Workflow):
            raise ExternalError(msg)



import inspect

from ..abstracts.event import Event
from ..abstracts.task import Task
from ..exceptions import ExternalError
from ..traits.with_timestamp import WithTimestamp
from ..traits.zenatonable import Zenatonable


class Wait(Task, Zenatonable, WithTimestamp):

    def __init__(self, event=None):
        super(Wait, self).__init__()
        if not self.valid_param(event):
            raise ExternalError(self.error)
        if event:
            self.event = event.__name__

    @property
    def error(self):
        return '{}: Invalid parameter - argument must be a zenaton.abstracts.event.Event subclass'.format(
            self.__class__.__name__)

    def handle(self):
        pass

    def valid_param(self, event):
        return not event or isinstance(event, str) or self.event_class(event)

    def event_class(self, event):
        return inspect.isclass(event) and issubclass(event, Event)
